# -*- coding: utf-8 -*-

import ast
import re
from pprint import pprint

import brainpy as bp
from brainpy.math.numpy.ast2numba import FindAllForLoop
from brainpy.math.numpy.ast2numba import _find_all_forloop
from brainpy.math.numpy.ast2numba import FuncTransformer
from brainpy.math.numpy.ast2numba import _jit_cls_func
from brainpy.tools import ast2code

bp.math.use_backend('numpy')


def test_find_self_data1():
  code = '''
  alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
  beta = 4.0 * bp.math.exp(-(V + 65) / 18)
  dmdt = alpha * (1 - m) - beta * m

  alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
  beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
  dhdt = alpha * (1 - h) - beta * h

  alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
  beta = 0.125 * bp.math.exp(-(V + 65) / 80)
  dndt = alpha * (1 - n) - beta * n

  I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
  I_K = (self.gK * n ** 4.0) * (V - self.EK)
  I_leak = self.gL * (V - self.EL)
  dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

  return dVdt, dmdt, dhdt, dndt
  '''

  arg = 'self'
  print(re.findall('\\b' + arg + '\\.[A-Za-z_][A-Za-z0-9_.]*\\b', code))


def test_transformer():
  code = '''
def update(self, _t, _dt):
  V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
  self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
  self.V[:] = V
  self.m[:] = m
  self.h[:] = h
  self.n[:] = n
  self.input[:] = 0.
  '''
  tree = ast.parse(code)

  transformer = FuncTransformer(func_name='self.integral', arg_to_append={'A_gNa': 'A.gNa'})
  new_tree = transformer.visit(tree)

  new_code = ast2code(new_tree)
  print(new_code)


def test_cls_func_hh1():
  class HH(bp.NeuGroup):
    target_backend = 'general'

    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = gNa
      self.gK = gK
      self.gL = gL
      self.V_th = V_th

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  hh = HH(10)

  r = _jit_cls_func(hh.update, show_code=True)
  pprint(r['func'])
  pprint('arguments:')
  pprint(r['arguments'])
  pprint('arg2call:')
  pprint(r['arg2call'])
  pprint('nodes:')
  pprint(r['nodes'])


def test_cls_func_hh2():
  class HH(bp.NeuGroup):
    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = bp.math.Variable(gNa)
      self.gK = gK
      self.gL = gL
      self.V_th = bp.math.Variable(V_th)

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  hh = HH(10)

  r = _jit_cls_func(hh.update, show_code=True)
  pprint(r['func'])
  pprint('arguments:')
  pprint(r['arguments'])
  pprint('arg2call:')
  pprint(r['arg2call'])
  pprint('nodes:')
  pprint(r['nodes'])


def test_cls_func_ampa1():
  class HH(bp.NeuGroup):
    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = bp.math.Variable(gNa)
      self.gK = gK
      self.gL = gL
      self.V_th = bp.math.Variable(V_th)

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  class AMPA_vec(bp.TwoEndConn):
    def __init__(self, pre, post, conn, delay=0., g_max=0.10, E=0., tau=2.0, **kwargs):
      super(AMPA_vec, self).__init__(pre=pre, post=post, **kwargs)

      # parameters
      self.g_max = g_max
      self.E = E
      self.tau = tau
      self.delay = delay

      # connections
      self.conn = conn(pre.size, post.size)
      self.pre_ids, self.post_ids = conn.requires('pre_ids', 'post_ids')
      self.size = len(self.pre_ids)

      # data
      self.s = bp.math.Variable(bp.math.zeros(self.size))
      self.g = self.register_constant_delay('g', size=self.size, delay=delay)

    @bp.odeint
    def int_s(self, s, t):
      return - s / self.tau

    def update(self, _t, _i):
      for i in range(self.size):
        pre_id = self.pre_ids[i]
        self.s[i] = self.int_s(self.s[i], _t)
        self.s[i] += self.pre.spike[pre_id]
        self.g.push(i, self.g_max * self.s[i])
        post_id = self.post_ids[i]
        self.post.input[post_id] -= self.g.pull(i) * (self.post.V[post_id] - self.E)

  hh = HH(10)
  ampa = AMPA_vec(pre=hh, post=hh, conn=bp.connect.All2All(), delay=10.)

  r = _jit_cls_func(ampa.update, show_code=True)
  pprint(r['func'])
  pprint('arguments:')
  pprint(r['arguments'])
  pprint('arg2call:')
  pprint(r['arg2call'])
  pprint('nodes:')
  pprint(r['nodes'])


def test_container1():
  class HH(bp.NeuGroup):
    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = bp.math.Variable(gNa)
      self.gK = gK
      self.gL = gL
      self.V_th = bp.math.Variable(V_th)

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  class AMPA_vec(bp.TwoEndConn):
    def __init__(self, pre, post, conn, delay=0., g_max=0.10, E=0., tau=2.0, **kwargs):
      super(AMPA_vec, self).__init__(pre=pre, post=post, **kwargs)

      # parameters
      self.g_max = g_max
      self.E = E
      self.tau = tau
      self.delay = delay

      # connections
      self.conn = conn(pre.size, post.size)
      self.pre_ids, self.post_ids = conn.requires('pre_ids', 'post_ids')
      self.size = len(self.pre_ids)

      # data
      self.s = bp.math.Variable(bp.math.zeros(self.size))
      self.g = self.register_constant_delay('g', size=self.size, delay=delay)

    @bp.odeint
    def int_s(self, s, t):
      return - s / self.tau

    def update(self, _t, _i):
      for i in range(self.size):
        pre_id = self.pre_ids[i]
        self.s[i] = self.int_s(self.s[i], _t)
        self.s[i] += self.pre.spike[pre_id]
        self.g.push(i, self.g_max * self.s[i])
        post_id = self.post_ids[i]
        self.post.input[post_id] -= self.g.pull(i) * (self.post.V[post_id] - self.E)

  hh = HH(10)
  ampa = AMPA_vec(pre=hh, post=hh, conn=bp.connect.All2All(), delay=10.)
  net = bp.Network(hh, ampa)

  r = _jit_cls_func(net.update, show_code=True)
  pprint(r['func'])
  pprint('arguments:')
  pprint(r['arguments'])
  pprint('arg2call:')
  pprint(r['arg2call'])
  pprint('nodes:')
  pprint(r['nodes'])


def test_hh1():
  class HH(bp.NeuGroup):
    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = gNa
      self.gK = gK
      self.gL = gL
      self.V_th = V_th

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  hh = HH(10)
  bp.math.jit(hh, show_code=True)


def test_hh2():
  class HH(bp.NeuGroup):
    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = bp.math.Variable(gNa)
      self.gK = gK
      self.gL = gL
      self.V_th = bp.math.Variable(V_th)

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  hh = HH(10)
  bp.math.jit(hh, show_code=True)


def test_hh_ampa_net1():
  class HH(bp.NeuGroup):
    def __init__(self, size, ENa=50., EK=-77., EL=-54.387,
                 C=1.0, gNa=120., gK=36., gL=0.03, V_th=20.,
                 **kwargs):
      # parameters
      self.ENa = ENa
      self.EK = EK
      self.EL = EL
      self.C = C
      self.gNa = bp.math.Variable(gNa)
      self.gK = gK
      self.gL = gL
      self.V_th = bp.math.Variable(V_th)

      # variables
      self.V = bp.math.Variable(bp.math.ones(size) * -65.)
      self.m = bp.math.Variable(bp.math.ones(size) * 0.5)
      self.h = bp.math.Variable(bp.math.ones(size) * 0.6)
      self.n = bp.math.Variable(bp.math.ones(size) * 0.32)
      self.spike = bp.math.Variable(bp.math.zeros(size, dtype=bool))
      self.input = bp.math.Variable(bp.math.zeros(size))

      super(HH, self).__init__(size=size, **kwargs)

    @bp.odeint(method='rk4')
    def integral(self, V, m, h, n, t, Iext):
      alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
      beta = 4.0 * bp.math.exp(-(V + 65) / 18)
      dmdt = alpha * (1 - m) - beta * m

      alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
      beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
      dhdt = alpha * (1 - h) - beta * h

      alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
      beta = 0.125 * bp.math.exp(-(V + 65) / 80)
      dndt = alpha * (1 - n) - beta * n

      I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
      I_K = (self.gK * n ** 4.0) * (V - self.EK)
      I_leak = self.gL * (V - self.EL)
      dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

      return dVdt, dmdt, dhdt, dndt

    def update(self, _t, _i):
      V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
      self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)
      self.V[:] = V
      self.m[:] = m
      self.h[:] = h
      self.n[:] = n
      self.input[:] = 0.

  class AMPA_vec(bp.TwoEndConn):
    def __init__(self, pre, post, conn, delay=0., g_max=0.10, E=0., tau=2.0, **kwargs):
      super(AMPA_vec, self).__init__(pre=pre, post=post, **kwargs)

      # parameters
      self.g_max = g_max
      self.E = E
      self.tau = tau
      self.delay = delay

      # connections
      self.conn = conn(pre.size, post.size)
      self.pre_ids, self.post_ids = conn.requires('pre_ids', 'post_ids')
      self.size = len(self.pre_ids)

      # data
      self.s = bp.math.Variable(bp.math.zeros(self.size))
      self.g = self.register_constant_delay('g', size=self.size, delay=delay)

    @bp.odeint
    def int_s(self, s, t):
      return - s / self.tau

    def update(self, _t, _i):
      for i in range(self.size):
        pre_id = self.pre_ids[i]
        self.s[i] = self.int_s(self.s[i], _t)
        self.s[i] += self.pre.spike[pre_id]
        self.g.push(i, self.g_max * self.s[i])
        post_id = self.post_ids[i]
        self.post.input[post_id] -= self.g.pull(i) * (self.post.V[post_id] - self.E)

  hh = HH(10)
  ampa = AMPA_vec(pre=hh, post=hh, conn=bp.connect.All2All(), delay=10.)
  net = bp.Network(hh, ampa)

  bp.math.jit(net, show_code=True)


def test_FindForLoop1():
  code = '''
for ch in self._update_channels:
  ch.update(_t, _dt)
for ch in self._current_channels:
  self.input += ch.update(_t, _dt)

# update variables
V = self.V + self.input / self.C / self.A * _dt
self.spike[:] = math.logical_and(V >= self.Vth, self.V < self.Vth)
self.V[:] = V
self.input[:] = 0.
  '''
  tree = ast.parse(code)

  finder = FindAllForLoop()
  finder.visit(tree)

  print()
  pprint(finder.for_target)
  pprint(finder.for_iter)
  pprint(finder.for_body)
  pprint(finder.for_body_str)

  print()
  pprint(_find_all_forloop(tree))


def test_FindForLoop2():
  code = '''
for step in self.child_steps.values():
  step(_t, _dt)
  '''

  tree = ast.parse(code)

  finder = FindAllForLoop()
  finder.visit(tree)

  print()
  pprint(finder.for_target)
  pprint(finder.for_iter)
  pprint(finder.for_body)
  pprint(finder.for_body_str)

  print()
  pprint(_find_all_forloop(tree))

def test_FindForLoop3():
  code = '''
for i in range(self.num):
  spike = False
  refractory = (_t - self.t_last_spike[i] <= self.t_refractory)
  if not refractory:
    V = self.integral(self.V[i], _t, self.input[i], dt=_dt)
    spike = (V >= self.V_th)
    if spike:
      V = self.V_reset
      self.t_last_spike[i] = _t
      refractory = True
    self.V[i] = V
  self.spike[i] = spike
  self.refractory[i] = refractory
self.input[:] = 0.
  '''
  tree = ast.parse(code)

  finder = FindAllForLoop()
  finder.visit(tree)

  print()
  pprint(finder.for_target)
  pprint(finder.for_iter)
  pprint(finder.for_body)
  pprint(finder.for_body_str)

  print()
  pprint(_find_all_forloop(tree))

def test_FindForLoop4():
  code = '''
self.pre_spike.push(self.pre.spike)
pre_spike = self.pre_spike.pull()

self.g[:] = self.integral(self.g, _t, dt=_dt)
for pre_id in range(self.pre.num):
  if pre_spike[pre_id]:
    start, end = self.pre_slice[pre_id]
    for post_id in self.post_ids[start: end]:
      self.g[post_id] += self.g_max

self.post.input[:] += self.output_current(self.g)
  '''
  tree = ast.parse(code)

  finder = FindAllForLoop()
  finder.visit(tree)

  print()
  pprint(finder.for_target)
  pprint(finder.for_iter)
  pprint(finder.for_body)
  pprint(finder.for_body_str)

  print()
  pprint(_find_all_forloop(tree))





# -*- coding: utf-8 -*-

import numba
import numpy
from numba.extending import overload


# @overload(numpy.shape)
# def shape_func(x):
#   if isinstance(x, (numba.types.Integer, numba.types.Float)):
#     def shape(x):
#       return (1,)
#
#     return shape
#   else:
#     return numpy.shape


@overload(numpy.clip)
def clip_func(x, x_min, x_max):
  def clip(x, x_min, x_max):
    x = numpy.maximum(x, x_min)
    x = numpy.minimum(x, x_max)
    return x

  return clip


@overload(numpy.squeeze)
def squeeze_func(a, axis=None):
  if isinstance(axis, numba.types.NoneType):
    def squeeze(a, axis=None):
      shape = []
      for s in a.shape:
        if s != 1:
          shape.append(s)
      return numpy.reshape(a, shape)

    return squeeze

  elif isinstance(axis, numba.types.Integer):
    def squeeze(a, axis=None):
      shape = []
      for i, s in enumerate(a.shape):
        if s != 1 or i != axis:
          shape.append(s)
      return numpy.reshape(a, shape)

    return squeeze

  else:
    def squeeze(a, axis=None):
      shape = []
      for i, s in enumerate(a.shape):
        if s != 1 or i not in axis:
          shape.append(s)
      return numpy.reshape(a, shape)

    return squeeze


def dsplit(): pass


def hsplit():  pass


def vsplit():  pass


def isreal(): pass


def isscalar(): pass


def meshgrid(): pass


def moveaxis(): pass


def ndim(): pass


def size(): pass


def take_along_axis(): pass


def tile(): pass



# -*- coding: utf-8 -*-

import io
import os
import re

from setuptools import find_packages
from setuptools import setup

# version
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'brainpy', '__init__.py'), 'r') as f:
  init_py = f.read()
version = re.search('__version__ = "(.*)"', init_py).groups()[0]

# obtain long description from README
with io.open(os.path.join(here, 'README.md'), 'r', encoding='utf-8') as f:
  README = f.read()

# setup
setup(
  name='brain-py',
  version=version,
  description='BrainPy: A flexible and extensible framework for brain modeling',
  long_description=README,
  long_description_content_type="text/markdown",
  author='Chaoming Wang',
  author_email='chao.brain@qq.com',
  packages=find_packages(exclude=['examples*', 'docs*', 'develop*', 'ode*']),
  python_requires='>=3.6',
  install_requires=[
    'numpy>=1.15',
    'matplotlib>=3.4',
  ],
  url='https://github.com/PKU-NIP-Lab/BrainPy',
  keywords='computational neuroscience, '
           'brain-inspired computation, '
           'dynamical systems, '
           'differential equations, '
           'numerical integration',
  classifiers=[
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Software Development :: Libraries',
  ]
)



# -*- coding: utf-8 -*-

from brainpy.simulation.brainobjects.base import DynamicalSystem

__all__ = [
  'Molecular'
]


class Molecular(DynamicalSystem):
  """Base class to model molecular objects.

  Parameters
  ----------

  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, name, **kwargs):
    super(Molecular, self).__init__(name=name, **kwargs)



# -*- coding: utf-8 -*-

import math as pm

from brainpy import errors
from brainpy import math as bm
from brainpy.simulation.brainobjects.base import DynamicalSystem
from brainpy.simulation.utils import size2len

__all__ = [
  'Delay',
  'ConstantDelay',
]


class Delay(DynamicalSystem):
  """Base class to model delay variables.

  Parameters
  ----------

  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, steps=('update',), name=None, monitors=None):
    super(Delay, self).__init__(steps=steps, monitors=monitors, name=name)

  def update(self, _t, _dt, **kwargs):
    raise NotImplementedError


class ConstantDelay(Delay):
  """Class used to model constant delay variables.

  For examples:

  >>> import brainpy as bp
  >>>
  >>> bp.ConstantDelay(size=10, delay=10.)
  >>> bp.ConstantDelay(size=100, delay=lambda: bp.math.random.randint(5, 10))
  >>> bp.ConstantDelay(size=100, delay=bp.math.random.random(100) * 4 + 10)

  Parameters
  ----------
  size : int, list of int, tuple of int
    The delay data size.
  delay : int, float, function, ndarray
    The delay time. With the unit of `dt`.
  num_batch : optional, int
    The batch size.
  steps : optional, tuple of str, tuple of function, dict of (str, function)
    The callable function, or a list of callable functions.
  monitors : optional, list, tuple, datastructures.Monitor
    Variables to monitor.
  name : optional, str
    The name of the dynamic system.
  """

  def __init__(self, size, delay, num_batch=None, dtype=None, dt=None, **kwargs):
    # dt
    self.dt = bm.get_dt() if dt is None else dt

    # num_batch
    self.num_batch = num_batch
    if num_batch is None:
      batch_size = ()
    else:
      assert isinstance(num_batch, int), 'Only support int for "num_batch"'
      batch_size = (num_batch,)

    # data size
    if isinstance(size, int): size = (size,)
    if not isinstance(size, (tuple, list)):
      raise errors.BrainPyError(f'"size" must a tuple/list of int, but we got {type(size)}: {size}')
    self.size = tuple(size)

    # data shape
    self.shape = self.size + batch_size

    # delay time length
    self.delay = delay

    # data and operations
    if isinstance(delay, (int, float)):  # uniform delay
      self.uniform_delay = True
      self.num_step = int(pm.ceil(delay / self.dt)) + 1
      self.out_idx = bm.Variable(bm.array([0]))
      self.in_idx = bm.Variable(bm.array([self.num_step - 1]))
      self.data = bm.Variable(bm.zeros((self.num_step,) + self.size + batch_size, dtype=dtype))

      self.push = self._push_for_uniform_delay
      self.pull = self._pull_for_uniform_delay

    else:  # non-uniform delay
      self.uniform_delay = False
      if not len(self.size) == 1:
        raise NotImplementedError(f'Currently, BrainPy only supports 1D heterogeneous '
                                  f'delays, while we got the heterogeneous delay with '
                                  f'{len(self.size)}-dimensions.')
      self.num = size2len(size)
      if callable(delay):  # like: "delay=lambda: np.random.randint(5, 10)"
        temp = bm.zeros(size)
        for i in range(size[0]): temp[i] = delay()
        delay = temp
      else:
        if bm.shape(delay) != self.size:
          raise errors.BrainPyError(f"The shape of the delay time size must be "
                                    f"the same with the delay data size. But we "
                                    f"got {bm.shape(delay)} != {self.size}")
      delay = bm.around(delay / self.dt)
      self.diag = bm.array(bm.arange(self.num), dtype=bm.int_)
      self.num_step = bm.array(delay, dtype=bm.int_) + 1
      self.in_idx = bm.Variable(self.num_step - 1)
      self.out_idx = bm.Variable(bm.zeros(self.num, dtype=bm.int_))
      self.data = bm.Variable(bm.zeros((self.num_step.max(),) + size + batch_size, dtype=dtype))

      self.push = self._push_for_nonuniform_delay
      self.pull = self._pull_for_nonuniform_delay

    super(ConstantDelay, self).__init__(**kwargs)

  def _pull_for_uniform_delay(self):
    """Pull delayed data for variables with the uniform delay."""
    return self.data[self.out_idx[0]]

  # def _pull_for_uniform_delay_batch(self):
  #   """Pull delayed data for variables with the uniform delay."""
  #   return self.data[:, self.out_idx[0]]

  def _push_for_uniform_delay(self, value):
    """Push the latest data to the delay bottom."""
    self.data[self.in_idx[0]] = value

  # def _push_for_uniform_delay_batch(self, value):
  #   """Push the latest data to the delay bottom."""
  #   self.data[:, self.in_idx[0]] = value

  def _pull_for_nonuniform_delay(self):
    """Pull delayed data for variables with the non-uniform delay."""
    return self.data[self.out_idx, self.diag]

  # def _pull_for_nonuniform_delay_batch(self):
  #   """Pull delayed data for variables with the non-uniform delay."""
  #   return self.data[:, self.out_idx, self.diag]

  def _push_for_nonuniform_delay(self, value):
    """Push the latest data to the delay bottom."""
    self.data[self.in_idx, self.diag] = value

  # def _push_for_nonuniform_delay_batch(self, value):
  #   """Push the latest data to the delay bottom."""
  #   self.data[:, self.in_idx, self.diag] = value

  def update(self, _t, _dt, **kwargs):
    """Update the delay index."""
    self.in_idx[:] = (self.in_idx + 1) % self.num_step
    self.out_idx[:] = (self.out_idx + 1) % self.num_step

  def reset(self):
    """Reset the variables."""
    self.in_idx[:] = self.num_step - 1
    self.out_idx[:] = 0
    self.data[:] = 0.



# -*- coding: utf-8 -*-

from brainpy import errors
from brainpy.base.base import Base

__all__ = [
  'Function',
]


class Function(Base):
  """The wrapper for Python functions.

  Parameters
  ----------
  f : function
    The function to wrap.
  nodes : list, tuple, dict
    The nodes in the defined function ``f``.
  name : optional, str
    The function name.
  """

  def __init__(self, f, nodes, name=None):
    # name
    self._f = f
    if name is None:
      name = self.unique_name(type=f.__name__ if hasattr(f, '__name__') else 'Function')

    # initialize
    super(Function, self).__init__(name=name)

    # nodes
    self._nodes = dict()
    if isinstance(nodes, Base):
      nodes = (nodes,)
    if isinstance(nodes, (tuple, list)):
      for i, node in enumerate(nodes):
        if not isinstance(node, Base):
          raise errors.BrainPyError(f'Must be an instance of {Base.__name__}, but we got {type(node)}.')
        self._nodes[f'_node{i}'] = node
    elif isinstance(nodes, dict):
      self._nodes.update(nodes)
    else:
      raise ValueError(f'Only support list/tuple/dict of {Base.__name__}, '
                       f'but we got {type(nodes)}: {nodes}')

  def nodes(self, method='absolute', _paths=None):
    if _paths is None:
      _paths = set()
    gather = self._nodes_in_container(self._nodes, method=method, _paths=_paths)
    gather.update(super(Function, self).nodes(method=method, _paths=_paths))
    return gather

  def __call__(self, *args, **kwargs):
    return self._f(*args, **kwargs)



# -*- coding: utf-8 -*-

from typing import List

CLASS_KEYWORDS: List[str] = ['self', 'cls']
SYSTEM_KEYWORDS: List[str] = ['_dt', '_t', '_i']


def set_class_keywords(*args):
  """Set the keywords for class specification.

  For example:

  >>> class A(object):
  >>>    def __init__(cls):
  >>>        pass
  >>>    def f(self, ):
  >>>        pass

  In this case, I use "cls" to denote the "self". So, I can set this by

  >>> set_class_keywords('cls', 'self')

  """
  global CLASS_KEYWORDS
  CLASS_KEYWORDS = list(args)


def get_class_keywords():
  return CLASS_KEYWORDS



# -*- coding: utf-8 -*-

import numpy as np

from brainpy import errors, math
from brainpy.simulation.brainobjects.neuron import NeuGroup


__all__ = [
  'SpikeTimeInput',
  'PoissonInput',
  'ConstantInput',
]


class SpikeTimeInput(NeuGroup):
  """The input neuron group characterized by spikes emitting at given times.

  >>> # Get 2 neurons, firing spikes at 10 ms and 20 ms.
  >>> SpikeTimeInput(2, times=[10, 20])
  >>> # or
  >>> # Get 2 neurons, the neuron 0 fires spikes at 10 ms and 20 ms.
  >>> SpikeTimeInput(2, times=[10, 20], indices=[0, 0])
  >>> # or
  >>> # Get 2 neurons, neuron 0 fires at 10 ms and 30 ms, neuron 1 fires at 20 ms.
  >>> SpikeTimeInput(2, times=[10, 20, 30], indices=[0, 1, 0])
  >>> # or
  >>> # Get 2 neurons; at 10 ms, neuron 0 fires; at 20 ms, neuron 0 and 1 fire;
  >>> # at 30 ms, neuron 1 fires.
  >>> SpikeTimeInput(2, times=[10, 20, 20, 30], indices=[0, 0, 1, 1])

  Parameters
  ----------
  size : int, tuple, list
      The neuron group geometry.
  indices : int, list, tuple
      The neuron indices at each time point to emit spikes.
  times : list, np.ndarray
      The time points which generate the spikes.
  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, size, times, indices, num_batch=None, need_sort=True, **kwargs):
    super(SpikeTimeInput, self).__init__(size=size, num_batch=num_batch, **kwargs)

    if len(indices) != len(times):
      raise errors.BrainPyError(f'The length of "indices" and "times" must be the same. '
                                f'However, we got {len(indices)} != {len(times)}.')

    # data about times and indices
    self.idx = 0
    self.times = math.asarray(times, dtype=math.float_)
    self.indices = np.asarray(indices, dtype=math.int_)
    self.num_times = len(times)
    if need_sort:
      sort_idx = np.argsort(times)
      self.indices = self.indices[sort_idx]
    self.spike = math.zeros(self.shape, dtype=bool)

  def update(self, _t, _i, **kwargs):
    self.spike[:] = False
    while self.idx < self.num_times and _t >= self.times[self.idx]:
      self.spike[self.indices[self.idx]] = 1.
      self.idx += 1


class PoissonInput(NeuGroup):
  """Poisson Neuron Group.

  Parameters
  ----------
  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, size, freqs, num_batch=None, seed=None, **kwargs):
    super(PoissonInput, self).__init__(size=size, num_batch=num_batch, **kwargs)

    self.freqs = freqs
    self.dt = math.get_dt() / 1000.
    self.size = (size,) if isinstance(size, int) else tuple(size)
    self.spike = math.Variable(math.zeros(self.num, dtype=bool))
    self.t_last_spike = math.Variable(math.ones(self.num) * -1e7)
    self.rng = math.random.RandomState(seed=seed)

  def update(self, _t, _i, **kwargs):
    self.spike[:] = self.rng.random(self.num) <= self.freqs * self.dt
    self.t_last_spike[:] = math.where(self.spike, _t, self.t_last_spike)


class ConstantInput(NeuGroup):
  pass



# -*- coding: utf-8 -*-

import inspect

from brainpy import errors, math
from brainpy.integrators import constants, utils
from brainpy.integrators.analysis_by_ast import separate_variables
from brainpy.integrators.ode import common

__all__ = [
  'general_rk_wrapper',
  'adaptive_rk_wrapper',
  'rk2_wrapper',
  'exp_euler_wrapper',
]

_f_kw = 'f'
_dt_kw = 'dt'


def general_rk_wrapper(f, show_code, dt, A, B, C, var_type, method):
  """Runge–Kutta methods for ordinary differential equation.

  For the system,

  .. backend::

      \frac{d y}{d t}=f(t, y)


  Explicit Runge-Kutta methods take the form

  .. backend::

      k_{i}=f\\left(t_{n}+c_{i}h,y_{n}+h\\sum _{j=1}^{s}a_{ij}k_{j}\\right) \\\\
      y_{n+1}=y_{n}+h \\sum_{i=1}^{s} b_{i} k_{i}

  Each method listed on this page is defined by its Butcher tableau,
  which puts the coefficients of the method in a table as follows:

  .. backend::

      \\begin{array}{c|cccc}
          c_{1} & a_{11} & a_{12} & \\ldots & a_{1 s} \\\\
          c_{2} & a_{21} & a_{22} & \\ldots & a_{2 s} \\\\
          \\vdots & \vdots & \vdots & \\ddots & \vdots \\\\
          c_{s} & a_{s 1} & a_{s 2} & \\ldots & a_{s s} \\\\
          \\hline & b_{1} & b_{2} & \\ldots & b_{s}
      \\end{array}

  Parameters
  ----------
  f : callable
      The derivative function.
  show_code : bool
      Whether show the formatted code.
  dt : float
      The numerical precision.
  A : tuple, list
      The A matrix in the Butcher tableau.
  B : tuple, list
      The B vector in the Butcher tableau.
  C : tuple, list
      The C vector in the Butcher tableau.

  Returns
  -------
  integral_func : callable
      The one-step numerical integration function.
  """
  class_kw, variables, parameters, arguments = utils.get_args(f)
  func_name = common.f_names(f)
  arguments = list(arguments) + [f'{_dt_kw}={dt}']

  keywords = {
    _f_kw: 'the derivative function',
    _dt_kw: 'the precision of numerical integration'
  }
  for v in variables:
    keywords[f'{v}_new'] = 'the intermediate value'
    for i in range(1, len(A) + 1):
      keywords[f'd{v}_k{i}'] = 'the intermediate value'
    for i in range(2, len(A) + 1):
      keywords[f'k{i}_{v}_arg'] = 'the intermediate value'
      keywords[f'k{i}_t_arg'] = 'the intermediate value'
  utils.check_kws(arguments, keywords)

  # code scope
  code_scope = {'f': f}

  # code lines
  code_lines = [f'def {func_name}({", ".join(arguments)}):']

  # step stage
  common.step(class_kw, variables, _dt_kw, A, C, code_lines, parameters)

  # variable update
  return_args = common.update(variables, _dt_kw, B, code_lines)

  # returns
  code_lines.append(f'  return {", ".join(return_args)}')

  # compilation
  return common.compile_and_assign_attrs(raw_func=dict(f=f),
                                         code_lines=code_lines,
                                         code_scope=code_scope,
                                         show_code=show_code,
                                         func_name=func_name,
                                         variables=variables,
                                         parameters=parameters,
                                         dt=dt,
                                         method=method,
                                         var_type=var_type)


def adaptive_rk_wrapper(f, dt, A, B1, B2, C, tol, adaptive, show_code, var_type, method):
  """Adaptive Runge-Kutta numerical method for ordinary differential equations.

  The embedded methods are designed to produce an estimate of the local
  truncation error of a single Runge-Kutta step, and as result, allow to
  control the error with adaptive stepsize. This is done by having two
  methods in the tableau, one with order p and one with order :backend:`p-1`.

  The lower-order step is given by

  .. backend::

      y^*_{n+1} = y_n + h\\sum_{i=1}^s b^*_i k_i,

  where the :backend:`k_{i}` are the same as for the higher order method. Then the error is

  .. backend::

      e_{n+1} = y_{n+1} - y^*_{n+1} = h\\sum_{i=1}^s (b_i - b^*_i) k_i,


  which is :backend:`O(h^{p})`. The Butcher Tableau for this kind of method is extended to
  give the values of :backend:`b_{i}^{*}`

  .. backend::

      \\begin{array}{c|cccc}
          c_1    & a_{11} & a_{12}& \\dots & a_{1s}\\\\
          c_2    & a_{21} & a_{22}& \\dots & a_{2s}\\\\
          \\vdots & \\vdots & \\vdots& \\ddots& \\vdots\\\\
          c_s    & a_{s1} & a_{s2}& \\dots & a_{ss} \\\\
      \\hline & b_1    & b_2   & \\dots & b_s\\\\
             & b_1^*    & b_2^*   & \\dots & b_s^*\\\\
      \\end{array}


  Parameters
  ----------
  f : callable
      The derivative function.
  show_code : bool
      Whether show the formatted code.
  dt : float
      The numerical precision.
  A : tuple, list
      The A matrix in the Butcher tableau.
  B1 : tuple, list
      The B1 vector in the Butcher tableau.
  B2 : tuple, list
      The B2 vector in the Butcher tableau.
  C : tuple, list
      The C vector in the Butcher tableau.
  adaptive : bool
  tol : float
  var_type : str

  Returns
  -------
  integral_func : callable
      The one-step numerical integration function.
  """
  if var_type not in constants.SUPPORTED_VAR_TYPE:
    raise errors.IntegratorError(f'"var_type" only supports {constants.SUPPORTED_VAR_TYPE}, not {var_type}.')

  class_kw, variables, parameters, arguments = utils.get_args(f)
  func_name = common.f_names(f)
  arguments = list(arguments) + [f'{_dt_kw}={dt}']

  keywords = {
    _f_kw: 'the derivative function',
    _dt_kw: 'the precision of numerical integration',
  }
  for v in variables:
    keywords[f'{v}_new'] = 'the intermediate value'
    for i in range(1, len(A) + 1):
      keywords[f'd{v}_k{i}'] = 'the intermediate value'
    for i in range(2, len(A) + 1):
      keywords[f'k{i}_{v}_arg'] = 'the intermediate value'
      keywords[f'k{i}_t_arg'] = 'the intermediate value'

  if adaptive:
    keywords['dt_new'] = 'the new numerical precision "dt"'
    keywords['tol'] = 'the tolerance for the local truncation error'
    keywords['error'] = 'the local truncation error'
    for v in variables:
      keywords[f'{v}_te'] = 'the local truncation error'
    # code scope
    code_scope = {_f_kw: f, 'tol': tol}
  else:
    # code scope
    code_scope = {_f_kw: f}
  utils.check_kws(arguments, keywords)

  # code lines
  code_lines = [f'def {func_name}({", ".join(arguments)}):']
  # stage steps
  common.step(class_kw, variables, _dt_kw, A, C, code_lines, parameters)
  # variable update
  return_args = common.update(variables, _dt_kw, B1, code_lines)

  # error adaptive item
  if adaptive:
    errors_ = []
    for v in variables:
      result = []
      for i, (b1, b2) in enumerate(zip(B1, B2)):
        if isinstance(b1, str):
          b1 = eval(b1)
        if isinstance(b2, str):
          b2 = eval(b2)
        diff = b1 - b2
        if diff != 0.:
          result.append(f'd{v}_k{i + 1} * {_dt_kw} * {diff}')
      if len(result) > 0:
        if var_type == constants.SCALAR_VAR:
          code_lines.append(f'  {v}_te = abs({" + ".join(result)})')
        else:
          code_lines.append(f'  {v}_te = sum(abs({" + ".join(result)}))')
        errors_.append(f'{v}_te')
    if len(errors_) > 0:
      code_lines.append(f'  error = {" + ".join(errors_)}')
      code_lines.append(f'  if error > tol:')
      code_lines.append(f'    {_dt_kw}_new = 0.9 * {_dt_kw} * (tol / error) ** 0.2')
      code_lines.append(f'  else:')
      code_lines.append(f'    {_dt_kw}_new = {_dt_kw}')
      return_args.append(f'{_dt_kw}_new')

  # returns
  code_lines.append(f'  return {", ".join(return_args)}')

  # compilation
  return common.compile_and_assign_attrs(raw_func=dict(f=f),
                                         code_lines=code_lines,
                                         code_scope=code_scope,
                                         show_code=show_code,
                                         func_name=func_name,
                                         variables=variables,
                                         parameters=parameters,
                                         dt=dt,
                                         method=method,
                                         var_type=var_type)


def rk2_wrapper(f, show_code, dt, beta, var_type):
  class_kw, variables, parameters, arguments = utils.get_args(f)
  func_name = common.f_names(f)
  arguments = list(arguments) + [f'{_dt_kw}={dt}']

  keywords = {
    _f_kw: 'the derivative function',
    _dt_kw: 'the precision of numerical integration',
    'beta': 'the parameters in RK2 method',
    '_k1': 'the parameters in RK2 method',
    '_k2': 'the parameters in RK2 method',
  }
  for v in variables:
    keywords[f'{v}_new'] = 'the intermediate value'
    for i in range(1, 3):
      keywords[f'd{v}_k{i}'] = 'the intermediate value'
    for i in range(2, 3):
      keywords[f'k{i}_{v}_arg'] = 'the intermediate value'
      keywords[f'k{i}_t_arg'] = 'the intermediate value'
  utils.check_kws(arguments, keywords)

  code_scope = {_f_kw: f,
                'beta': beta,
                '_k1': 1 - 1 / (2 * beta),
                '_k2': 1 / (2 * beta)}
  code_lines = [f'def {func_name}({", ".join(arguments)}):']
  # k1
  k1_args = variables + parameters
  k1_vars_d = [f'd{v}_k1' for v in variables]
  code_lines.append(f'  {", ".join(k1_vars_d)} = f({", ".join(class_kw + k1_args)})')
  # k2
  k2_args = [f'{v} + d{v}_k1 * dt * beta' for v in variables]
  k2_args.append('t + dt * beta')
  k2_args.extend(parameters[1:])
  k2_vars_d = [f'd{v}_k2' for v in variables]
  code_lines.append(f'  {", ".join(k2_vars_d)} = f({", ".join(class_kw + k2_args)})')
  # returns
  for v, k1, k2 in zip(variables, k1_vars_d, k2_vars_d):
    code_lines.append(f'  {v}_new = {v} + ({k1} * _k1 + {k2} * _k2) * dt')
  return_vars = [f'{v}_new' for v in variables]
  code_lines.append(f'  return {", ".join(return_vars)}')

  return common.compile_and_assign_attrs(raw_func=dict(f=f),
                                         code_lines=code_lines,
                                         code_scope=code_scope,
                                         show_code=show_code,
                                         func_name=func_name,
                                         variables=variables,
                                         parameters=parameters,
                                         dt=dt,
                                         method='rk2',
                                         var_type=var_type)


def exp_euler_wrapper(f, show_code, dt, var_type):
  try:
    import sympy
    from brainpy.integrators import analysis_by_sympy
  except ModuleNotFoundError:
    raise errors.PackageMissingError('SymPy must be installed when '
                                     'using exponential euler methods.')

  if var_type == constants.SYSTEM_VAR:
    raise errors.IntegratorError(f'Exponential Euler method do not '
                                 f'support {var_type} variable type.')

  class_kw, variables, parameters, arguments = utils.get_args(f)
  func_name = common.f_names(f)
  arguments = list(arguments) + [f'{_dt_kw}={dt}']
  keywords = {
    _f_kw: 'the derivative function',
    _dt_kw: 'the precision of numerical integration',
    'exp': 'the exponential function',
    'math': 'the math module',
  }
  for v in variables:
    keywords[f'{v}_new'] = 'the intermediate value'
  utils.check_kws(arguments, keywords)

  code_lines = [f'def {func_name}({", ".join(arguments)}):']

  # code scope
  closure_vars = inspect.getclosurevars(f)
  code_scope = dict(closure_vars.nonlocals)
  code_scope.update(dict(closure_vars.globals))
  code_scope[_f_kw] = f
  code_scope['math'] = math

  analysis = separate_variables(f)
  variables_for_returns = analysis['variables_for_returns']
  expressions_for_returns = analysis['expressions_for_returns']
  for vi, (key, vars) in enumerate(variables_for_returns.items()):
    # separate variables
    sd_variables = []
    for v in vars:
      if len(v) > 1:
        raise ValueError('Cannot analyze multi-assignment code line.')
      sd_variables.append(v[0])
    expressions = expressions_for_returns[key]
    var_name = variables[vi]
    diff_eq = analysis_by_sympy.SingleDiffEq(var_name=var_name,
                                             variables=sd_variables,
                                             expressions=expressions,
                                             derivative_expr=key,
                                             scope=code_scope,
                                             func_name=func_name)

    f_expressions = diff_eq.get_f_expressions(substitute_vars=diff_eq.var_name)

    # code lines
    code_lines.extend([f"  {str(expr)}" for expr in f_expressions[:-1]])

    # get the linear system using sympy
    f_res = f_expressions[-1]
    df_expr = analysis_by_sympy.str2sympy(f_res.code).expr.expand()
    s_df = sympy.Symbol(f"{f_res.var_name}")
    code_lines.append(f'  {s_df.name} = {analysis_by_sympy.sympy2str(df_expr)}')
    var = sympy.Symbol(diff_eq.var_name, real=True)

    # get df part
    s_linear = sympy.Symbol(f'_{diff_eq.var_name}_linear')
    s_linear_exp = sympy.Symbol(f'_{diff_eq.var_name}_linear_exp')
    s_df_part = sympy.Symbol(f'_{diff_eq.var_name}_df_part')
    if df_expr.has(var):
      # linear
      linear = sympy.collect(df_expr, var, evaluate=False)[var]
      code_lines.append(f'  {s_linear.name} = {analysis_by_sympy.sympy2str(linear)}')
      # linear exponential
      # linear_exp = sympy.exp(linear * dt)
      # code_lines.append(f'  {s_linear_exp.name} = {analysis_by_sympy.sympy2str(linear_exp)}')
      code_lines.append(f'  {s_linear_exp.name} = math.exp({s_linear.name} * {_dt_kw})')
      # df part
      df_part = (s_linear_exp - 1) / s_linear * s_df
      code_lines.append(f'  {s_df_part.name} = {analysis_by_sympy.sympy2str(df_part)}')

    else:
      # # linear exponential
      # code_lines.append(f'  {s_linear_exp.name} = {_dt_kw} ** 0.5')
      # df part
      code_lines.append(f'  {s_df_part.name} = {s_df.name} * {_dt_kw}')

    # update expression
    update = var + s_df_part

    # The actual update step
    code_lines.append(f'  {diff_eq.var_name}_new = {analysis_by_sympy.sympy2str(update)}')
    code_lines.append('')

  code_lines.append(f'  return {", ".join([f"{v}_new" for v in variables])}')
  return common.compile_and_assign_attrs(raw_func=dict(f=f),
                                         code_lines=code_lines,
                                         code_scope=code_scope,
                                         show_code=show_code,
                                         func_name=func_name,
                                         variables=variables,
                                         parameters=parameters,
                                         method='exponential_euler',
                                         dt=dt,
                                         var_type=var_type)



# -*- coding: utf-8 -*-


import numpy as np
import brainpy.math as bm
from brainpy.simulation.brainobjects.delays import ConstantDelay


def test_constant_delay_uniform_no_batch1():
  print()

  for bk in ['jax', 'numpy']:
    bm.use_backend(bk)
    cd = ConstantDelay(size=10, delay=2, dt=0.1)
    for i in range(cd.num_step):
      cd.push(bm.ones(cd.shape) * i)
      cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    a = cd.pull()
    print(a)
    print(type(a))


def test_constant_delay_uniform_batch1():
  print()

  for bk in ['jax', 'numpy']:
    bm.use_backend(bk)
    cd = ConstantDelay(size=10, delay=2, dt=0.1, num_batch=2)
    for i in range(cd.num_step):
      cd.push(bm.ones(cd.shape) * i)
      cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    a = cd.pull()
    print(a)
    print(type(a))


def test_constant_delay_nonuniform_no_batch1():
  print()

  rng = np.random.RandomState(1234)
  delays = rng.random(10) * 3 + 0.2

  for bk in ['jax', 'numpy']:
    bm.use_backend(bk)
    cd = ConstantDelay(size=10, delay=delays, dt=0.1)
    for i in range(cd.num_step.max()):
      cd.push(bm.ones(cd.shape) * i)
      cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    a = cd.pull()
    print(a)
    print(type(a))


def test_constant_delay_nonuniform_batch1():
  print()

  rng = np.random.RandomState(1234)
  delays = rng.random(10) * 3 + 0.2

  for bk in ['jax', 'numpy']:
    bm.use_backend(bk)
    cd = ConstantDelay(size=10, delay=delays, dt=0.1, num_batch=2)
    for i in range(cd.num_step.max()):
      cd.push(bm.ones(cd.shape) * i)
      cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    print(cd.pull())
    cd.update(0, 0)
    a = cd.pull()
    print(a)
    print(type(a))




# -*- coding: utf-8 -*-

from brainpy import errors, math
from brainpy.simulation.brainobjects.base import DynamicalSystem
from brainpy.simulation.brainobjects.neuron import NeuGroup
from brainpy.simulation.connect import TwoEndConnector, MatConn, IJConn

__all__ = [
  'TwoEndConn',
]


class TwoEndConn(DynamicalSystem):
  """Base class to model two-end synaptic connections.

  Parameters
  ----------
  steps : function, list of function, tuple of function, dict of (str, function), optional
      The step functions.
  pre : NeuGroup
      Pre-synaptic neuron group.
  post : NeuGroup
      Post-synaptic neuron group.
  conn : math.ndarray, dict of (str, math.ndarray), TwoEndConnector
      The connection method between pre- and post-synaptic groups.
  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  show_code : bool
      Whether show the formatted code.
  """

  def __init__(self, pre, post, conn=None, name=None, steps=('update',), **kwargs):
    # pre or post neuron group
    # ------------------------
    if not isinstance(pre, NeuGroup):
      raise errors.BrainPyError('"pre" must be an instance of NeuGroup.')
    if not isinstance(post, NeuGroup):
      raise errors.BrainPyError('"post" must be an instance of NeuGroup.')
    if pre.num_batch != post.num_batch:
      raise errors.BrainPyError('pre.num_batch != post.num_batch')
    self.pre = pre
    self.post = post
    assert pre.num_batch is None, 'currently, synaptic computing does not support batch'

    # connectivity
    # ------------
    if isinstance(conn, TwoEndConnector):
      self.conn = conn(pre.size, post.size)
    elif isinstance(conn, math.ndarray):
      if (pre.num, post.num) != conn.shape:
        raise errors.BrainPyError(f'"conn" is provided as a matrix, and it is expected '
                                  f'to be an array with shape of (pre.num, post.num) = '
                                  f'{(pre.num, post.num)}, however we got {conn.shape}')
      self.conn = MatConn(conn_mat=conn)
    elif isinstance(conn, dict):
      if not ('i' in conn and 'j' in conn):
        raise errors.BrainPyError(f'"conn" is provided as a dict, and it is expected to '
                                  f'be a dictionary with "i" and "j" specification, '
                                  f'however we got {conn}')
      self.conn = IJConn(i=conn['i'], j=conn['j'])
    elif conn is None:
      self.conn = conn
    else:
      raise errors.BrainPyError(f'Unknown "conn" type: {conn}')

    # initialize
    # ----------
    super(TwoEndConn, self).__init__(steps=steps, name=name, **kwargs)



# -*- coding: utf-8 -*-


import brainpy as bp

def test1():
  assert bp.Network().name.startswith('Network')
  assert bp.Container().name.startswith('Container')
  assert bp.NeuGroup(10).name.startswith('NeuGroup')





# -*- coding: utf-8 -*-

from typing import cast

import numpy as np
try:
  import numba as nb
except ModuleNotFoundError:
  nb = None


__all__ = [
  'ndarray',
  'Variable',
  'TrainVar',
  'Parameter',
]

ndarray = np.ndarray


class Variable(np.ndarray):
  """Variable.

  """
  def __new__(cls, value, type='', replicate=None):
    if not isinstance(value, ndarray):
      arr_value = np.asarray(value)
    else:
      arr_value = value
    obj = arr_value.view(cls)
    if arr_value.dtype == np.dtype('O'):
      if isinstance(value, (list, tuple)):
        if len(value) > 0 and isinstance(value[0], ndarray) and nb is not None:
          value2 = nb.typed.List()
          for v in value:
            value2.append(v)
          value = value2
      obj.value = value
    else:
      obj.value = arr_value
    obj.type = type
    obj.replicate = replicate
    return obj

  def __array_finalize__(self, obj):
    if obj is None: return
    self.replicate = getattr(obj, 'replicate', None)
    self.value = getattr(obj, 'value', None)
    self.type = getattr(obj, 'type', None)

  def issametype(self, other):
    if self.type:
      return not isinstance(other, Variable)
    else:
      if not isinstance(other, Variable):
        return False
      else:
        return other.type == self.type


class TrainVar(Variable):
  """Trainable Variable.

  """
  __slots__ = ()

  def __new__(cls, value, replicate=None):
    return cast(TrainVar, super().__new__(cls, value=value, type='train', replicate=replicate))


class Parameter(Variable):
  """Parameter.

  """
  __slots__ = ()

  def __new__(cls, value, replicate=None):
    return cast(TrainVar, super().__new__(cls, value=value, type='param', replicate=replicate))



# -*- coding: utf-8 -*-


import numpy as np

__all__ = [
  'cluster_coefficient',
]


def cluster_coefficient(conn_mat):
  ccs = []
  nodes = np.arange(conn_mat.shape[0])
  for node in nodes:
    triangles = 0
    neighbor_pairs = 0
    neighbors = np.where(conn_mat[node])[0]
    for neighbor1 in neighbors:
      for neighbor2 in neighbors:
        if neighbor1 == neighbor2:
          continue
        neighbor_pairs += 1
        if conn_mat[neighbor1, neighbor2]:
          triangles += 1
    ccs.append(triangles / neighbor_pairs)
  return sum(ccs) / len(ccs)



# -*- coding: utf-8 -*-

import pytest

from brainpy import errors
from brainpy.integrators import ode
from brainpy.integrators import odeint


def test_euler():
    print('Test Euler method:')
    print()
    odeint(method='euler', show_code=True, f=lambda v, t, p: t)

    with pytest.raises(errors.CodeError):
        odeint(method='euler', show_code=True, f=lambda f, t, dt: t)

    with pytest.raises(errors.CodeError):
        odeint(method='euler', show_code=True, f=lambda v, t, dt: t)

    with pytest.raises(errors.CodeError):
        odeint(method='euler', show_code=True, f=lambda v, t, v_new: t)

    with pytest.raises(errors.CodeError):
        odeint(method='euler', show_code=True, f=lambda v, t, dv_k1: t)

    print('-' * 40)


def test_order2_rk():
    for method in ['heun2', 'midpoint', 'ralston2']:

        print(f'Test {method} method:')
        print()
        odeint(method=method, show_code=True, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k2_t_arg: t)

        print('-' * 40)


def test_rk2():
    method = 'rk2'

    print(f'Test {method} method:')
    print()
    odeint(method=method, show_code=True, f=lambda v, t, p: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda f, t, dt: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, dt: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, v_new: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, dv_k1: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, dv_k2: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, k2_v_arg: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, k2_t_arg: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, beta: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, _k1: t)

    with pytest.raises(errors.CodeError):
        odeint(method=method, show_code=True, f=lambda v, t, _k2: t)

    print('-' * 40)



def test_order3_rk():
    for method in ['rk3', 'heun3', 'ralston3', 'ssprk3']:

        print(f'Test {method} method:')
        print()
        odeint(method=method, show_code=True, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k3: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k3_t_arg: t)

        print('-' * 40)


def test_order4_rk():
    for method in ['rk4', 'ralston4', 'rk4_38rule']:

        print(f'Test {method} method:')
        print()
        odeint(method=method, show_code=True, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k3: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, dv_k4: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k3_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k4_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, f=lambda v, t, k4_t_arg: t)

        print('-' * 40)


# if __name__ == '__main__':
#     test_euler()
#     test_order2_rk()
#     test_rk2()
#     test_order3_rk()
#     test_order4_rk()



# -*- coding: utf-8 -*-

import abc
from typing import Union, List, Tuple

from brainpy import errors, tools
from brainpy.simulation.connect import formatter

__all__ = [
  'CONN_MAT',
  'PRE_IDS', 'POST_IDS',
  'PRE2POST', 'POST2PRE',
  'PRE2SYN', 'POST2SYN',
  'PRE_SLICE', 'POST_SLICE',
  'SUPPORTED_SYN_STRUCTURE',

  'PROVIDE_MAT', 'PROVIDE_IJ',

  'Connector', 'TwoEndConnector', 'OneEndConnector',
]

CONN_MAT = 'conn_mat'
PRE_IDS = 'pre_ids'
POST_IDS = 'post_ids'
PRE2POST = 'pre2post'
POST2PRE = 'post2pre'
PRE2SYN = 'pre2syn'
POST2SYN = 'post2syn'
PRE_SLICE = 'pre_slice'
POST_SLICE = 'post_slice'

SUPPORTED_SYN_STRUCTURE = [CONN_MAT,
                           PRE_IDS, POST_IDS,
                           PRE2POST, POST2PRE,
                           PRE2SYN, POST2SYN,
                           PRE_SLICE, POST_SLICE]

PROVIDE_MAT = 'mat'
PROVIDE_IJ = 'ij'


class Connector(abc.ABC):
  """Base Synaptical Connector Class."""
  pass


class TwoEndConnector(Connector):
  """Synaptical connector to build synapse connections between two neuron groups."""

  def __call__(self, pre_size, post_size):
    """Create the concrete connections between two end objects.

    Parameters
    ----------
    pre_size : int, tuple of int, list of int
        The size of the pre-synaptic group.
    post_size : int, tuple of int, list of int
        The size of the post-synaptic group.

    Returns
    -------
    conn : TwoEndConnector
        Return the self.
    """
    if isinstance(pre_size, int): pre_size = (pre_size,)
    pre_size = tuple(pre_size)
    if isinstance(post_size, int): post_size = (post_size,)
    post_size = tuple(post_size)
    self.pre_size, self.post_size = pre_size, post_size
    self.pre_num = tools.size2num(self.pre_size)
    self.post_num = tools.size2num(self.post_size)
    return self

  def check(self, structures: Union[Tuple, List, str]):
    if (not hasattr(self, 'pre_size')) or (not hasattr(self, 'post_size')):
      raise errors.BrainPyError(f'Please call "__call__" first to gather the size of the '
                                f'pre-synaptic and post-synaptic neuron groups for: {str(self)}')

    # check synaptic structures
    if isinstance(structures, str):
      structures = (structures,)
    if 'pre_slice' in structures and 'post_slice' in structures:
      raise errors.BrainPyError('Cannot use "pre_slice" and "post_slice" simultaneously. \n'
                                'We recommend you use "pre_slice + post2syn" or "post_slice + pre2syn".')
    self.structures = structures

    # get synaptic structures
    for n in structures:
      if n not in SUPPORTED_SYN_STRUCTURE:
        raise ValueError(f'Unknown synapse structure {n}. We only '
                         f'support {SUPPORTED_SYN_STRUCTURE}.')

    # provide what synaptic structure?
    if len(structures) == 0 or CONN_MAT in structures:
      return PROVIDE_MAT
    else:
      return PROVIDE_IJ

  def returns(self, mat=None, ij=None):
    if not hasattr(self, 'structures'):
      raise errors.BrainPyError(f'Please call "self.check" first to get synaptic '
                                f'structures required. Error in {str(self)}')

    if len(self.structures) == 1 and self.structures[0] == CONN_MAT:
      if mat is None:
        mat = formatter.ij2mat(i=ij[0], j=ij[1], num_pre=self.pre_num, num_post=self.post_num)
      return mat
    else:
      all_data = dict()

      # check 'ij'
      if ij is None:
        if mat is None:
          raise errors.BrainPyError(f'"mat" and "ij" are both none, please provide at least one of them.')
        ij = formatter.mat2ij(mat)
      all_data[PRE_IDS] = ij[0]
      all_data[POST_IDS] = ij[1]

      # check 'mat'
      if CONN_MAT in self.structures:
        if mat is None:
          mat = formatter.ij2mat(i=ij[0], j=ij[1], num_pre=self.pre_num, num_post=self.post_num)
      all_data[CONN_MAT] = mat

      # names of the needed structures
      if PRE_SLICE in self.structures:
        all_data[PRE_IDS], all_data[POST_IDS], all_data[PRE_SLICE] = \
          formatter.pre_slice(i=ij[0], j=ij[1], num_pre=self.pre_num)
      elif POST_SLICE in self.structures:
        all_data[PRE_IDS], all_data[POST_IDS], all_data[POST_SLICE] = \
          formatter.post_slice(i=ij[0], j=ij[1], num_post=self.post_num)
      for n in self.structures:
        if n in [PRE_SLICE, POST_SLICE, PRE_IDS, POST_IDS, CONN_MAT]:
          continue
        elif n == PRE2POST:
          all_data[PRE2POST] = formatter.pre2post(i=ij[0], j=ij[1], num_pre=self.pre_num)
        elif n == PRE2SYN:
          all_data[PRE2SYN] = formatter.pre2syn(i=ij[0], num_pre=self.pre_num)
        elif n == POST2PRE:
          all_data[POST2PRE] = formatter.post2pre(i=ij[0], j=ij[1], num_post=self.post_num)
        elif n == POST2SYN:
          all_data[POST2SYN] = formatter.post2syn(j=ij[1], num_post=self.post_num)
        else:
          raise errors.BrainPyError

      # data of the needed structures
      if len(self.structures) == 1:
        return all_data[self.structures[0]]
      else:
        return tuple([all_data[n] for n in self.structures])

  def requires(self, *structures):
    return self.require(*structures)

  def require(self, *structures):
    raise NotImplementedError


class OneEndConnector(TwoEndConnector):
  """Synaptical connector to build synapse connections within a population of neurons."""

  def __call__(self, pre_size, post_size=None):
    if post_size is None: post_size = pre_size
    else: assert pre_size == post_size
    if isinstance(pre_size, int): pre_size = (pre_size,)
    pre_size = tuple(pre_size)
    if isinstance(post_size, int): post_size = (post_size,)
    post_size = tuple(post_size)
    self.pre_size, self.post_size = pre_size, post_size
    self.pre_num = tools.size2num(self.pre_size)
    self.post_num = tools.size2num(self.post_size)
    return self




# -*- coding: utf-8 -*-

"""
The base module for whole BrainPy ecosystem.

This module provides the most fundamental class ``Base``,
and its associated helper class ``Collector``.

For each instance of "Base" class, users can retrieve all
the variables (or trainable variables), integrators, and nodes.

This module also provides a ``Function`` class to wrap user-defined
functions. In each function, maybe several nodes are used, and
users can initialize a ``Function`` by providing the nodes used
in the function. Unfortunately, ``Function`` class does not have
the ability to gather nodes automatically.
"""

from brainpy.base.base import *
from brainpy.base.collector import *
from brainpy.base.function import *
from brainpy.base.io import *




# -*- coding: utf-8 -*-

import numpy
import numpy.random

from brainpy import tools

try:
  import numba
except ModuleNotFoundError:
  numba = None

__all__ = [
  'seed', 'RandomState',
  'rand', 'randint', 'randn', 'random', 'random_sample',
  'ranf', 'sample', 'choice', 'permutation', 'shuffle',
  'beta', 'exponential', 'gamma', 'gumbel', 'laplace', 'logistic', 'normal', 'pareto',
  'poisson', 'standard_cauchy', 'standard_exponential', 'standard_gamma',
  'standard_normal', 'standard_t', 'uniform', 'bernoulli', 'truncated_normal',
  # 'binomial', 'chisquare', 'f', 'geometric', 'hypergeometric', 'lognormal',
  # 'logseries', 'multinomial', 'negative_binomial', 'power', 'rayleigh',
  # 'triangular', 'vonmises', 'wald', 'weibull',
]

# def RandomState(seed=None):
#   if seed:
#     if numba:
#       numba_seed(seed)
#     numpy.random.seed(seed)
#   return numpy.random

RandomState = numpy.random.RandomState
rand = numpy.random.rand
randint = numpy.random.randint
randn = numpy.random.randn
random = numpy.random.random
random_sample = numpy.random.random_sample
ranf = numpy.random.ranf
sample = numpy.random.sample
choice = numpy.random.choice
permutation = numpy.random.permutation
shuffle = numpy.random.shuffle
beta = numpy.random.beta
exponential = numpy.random.exponential
gamma = numpy.random.gamma
gumbel = numpy.random.gumbel
laplace = numpy.random.laplace
logistic = numpy.random.logistic
normal = numpy.random.normal
pareto = numpy.random.pareto
poisson = numpy.random.poisson
standard_cauchy = numpy.random.standard_cauchy
standard_exponential = numpy.random.standard_exponential
standard_gamma = numpy.random.standard_gamma
standard_normal = numpy.random.standard_normal
standard_t = numpy.random.standard_t
uniform = numpy.random.uniform


def bernoulli(p, size=None):
  return numpy.random.binomial(1, p=p, size=size)


def truncated_normal():
  raise NotImplementedError


@tools.numba_jit
def numba_seed(seed=None):
  numpy.random.seed(seed)


def seed(seed=None):
  if seed is not None:
    numpy.random.seed(seed)
    if numba:
      numba_seed(seed)



# -*- coding: utf-8 -*-

import inspect

from brainpy import errors
from brainpy.simulation.brainobjects.base import DynamicalSystem

__all__ = [
  'Sequential',
]


def _check_args(args):
  if args is None:
    return tuple()
  elif isinstance(args, tuple):
    return args
  else:
    return (args,)


class Sequential(DynamicalSystem):
  """Basic sequential object to control data flow.

  Parameters
  ----------
  arg_ds
    The modules without name specifications.
  name : str, optional
    The name of the sequential module.
  kwarg_ds
    The modules with name specifications.
  """

  def __init__(self, *arg_ds, monitors=None, name=None, **kwarg_ds):
    super(Sequential, self).__init__(monitors=monitors, name=name)

    self.child_ds = dict()
    # check "args"
    for ds in arg_ds:
      if not isinstance(ds, DynamicalSystem):
        raise errors.BrainPyError(f'Only support {DynamicalSystem.__name__}, '
                                  f'but we got {type(ds)}: {str(ds)}.')
      self.child_ds[ds.name] = ds

    # check "kwargs"
    for key, ds in kwarg_ds.items():
      if not isinstance(ds, DynamicalSystem):
        raise errors.BrainPyError(f'Only support {DynamicalSystem.__name__}, '
                                  f'but we got {type(ds)}: {str(ds)}.')
      self.child_ds[key] = ds

    # all update functions
    self._return_kwargs = ['kwargs' in inspect.signature(ds.update).parameters.keys()
                           for ds in self.child_ds.values()]

  def _check_kwargs(self, i, kwargs):
    return kwargs if self._return_kwargs[i] else dict()

  def update(self, *args, **kwargs):
    """Functional call.

    Parameters
    ----------
    args : list, tuple
      The *args arguments.
    kwargs : dict
      The config arguments. The configuration used across modules.
      If the "__call__" function in submodule receives "config" arguments,
      This "config" parameter will be passed into this function.
    """
    ds = list(self.child_ds.values())
    # first layer
    args = ds[0].update(*args, **self._check_kwargs(0, kwargs))
    # other layers
    for i in range(1, len(self.child_ds)):
      args = ds[i].update(*_check_args(args=args), **self._check_kwargs(i, kwargs))
    return args

  def nodes(self, method='absolute', _paths=None):
    if _paths is None:
      _paths = set()
    gather = self._nodes_in_container(self.child_ds, method=method, _paths=_paths)
    gather.update(super(Sequential, self).nodes(method=method))
    return gather

  def init(self, num_batch=1, **kwargs):
    for ds in self.child_ds.values():
      ds.init(num_batch=num_batch, **kwargs)
